/* AlignmentViewer.java
 *
 * created: Mon Jul 12 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999,2000  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/AlignmentViewer.java,v 1.10 2000/05/11 11:08:02 kmr Exp $
 */

package diana.components;

import diana.*;
import diana.sequence.*;
import uk.ac.sanger.pathogens.OutOfRangeException;

import java.awt.*;
import java.awt.event.*;

/**
 *  This component shows an alignment of two sequences using the data from a
 *  ComparisonData object.
 *
 *  @author Kim Rutherford
 *  @version $Id: AlignmentViewer.java,v 1.10 2000/05/11 11:08:02 kmr Exp $
 **/

public class AlignmentViewer extends Panel
    implements SequenceChangeListener {
  /**
   *  Create a new AlignmentViewer for the given entries.
   *  @param comparison_data Provides the AlignMatch objects that will be
   *    displayed.
   **/
  public AlignmentViewer (final Comparator owner,
                          final ComparisonData comparison_data) {
    this.owner              = owner;
    this.comparison_data    = comparison_data;

    alignment_data_array = comparison_data.getAlignments ();

    owner.getTopEntryGroup ().getBases ().addSequenceChangeListener (this, 0);
    owner.getBottomEntryGroup ().getBases ().addSequenceChangeListener (this, 0);

    top_forward_strand = getTopForwardStrand ();
    bottom_forward_strand = getBottomForwardStrand ();

    setLayout(new BorderLayout ());

    score_colours = makeColours ();

    createCanvas ();

    scroll_bar =
      new Scrollbar (Scrollbar.VERTICAL, 1, 1, 1, 1000);

    scroll_bar.setBlockIncrement (10);

    scroll_bar.addAdjustmentListener (new AdjustmentListener () {
      public void adjustmentValueChanged(AdjustmentEvent e) {
        repaintCanvas ();
      }
    });

    add (scroll_bar, "East");
  }

  /**
   *  This method tells this AlignmentViewer component where the top sequence
   *  is now.
   **/
  public void setTopSequencePosition (final DisplayAdjustmentEvent event) {
    last_top_event = event;
    repaintCanvas ();
  }

  /**
   *  This method tells this AlignmentViewer component where the bottom
   *  sequence is now.
   **/
  public void setBottomSequencePosition (final DisplayAdjustmentEvent event) {
    last_bottom_event = event;
    repaintCanvas ();
  }

  /**
   *  Implementation of the SequenceChangeListener interface.  The display is
   *  redrawn if there is an event.
   **/
  public void sequenceChanged (final SequenceChangeEvent event) {
    repaintCanvas ();
  }

  /**
   *  Create the canvas object and add it.
   **/
  private void createCanvas () {
    canvas = new Canvas ()
      {
      /**
       *  Set the offscreen buffer to null as part of invalidation.
       **/
      public void invalidate() {
        super.invalidate();
        offscreen = null;
      }

      /**
       *  Override update to *not* erase the background before painting
       */
      public void update(final Graphics g) {
        paint(g);
      }

      /**
       *  Paint the canvas.
       */
      public void paint(final Graphics g) {
        paintCanvas (g);
      }

    };

    add (canvas, "Center");

    canvas.addMouseListener (new MouseAdapter () {
      public void mousePressed (final MouseEvent event) {
        handleCanvasMousePress (event);

        if (event.isPopupTrigger () || event.isMetaDown ()) {

          final PopupMenu popup = new PopupMenu ();

          final MenuItem cutoffs_item = new MenuItem ("Set Cutoffs ...");

          popup.add (cutoffs_item);

          cutoffs_item.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent _) {
              final ScoreChangeListener minimum_listener =
                new ScoreChangeListener () {
                  public void scoreChanged (final ScoreChangeEvent event) {
                    minimum_score = event.getValue ();
                    repaintCanvas ();
                  }
                };

              final ScoreChangeListener maximum_listener =
                new ScoreChangeListener () {
                  public void scoreChanged (final ScoreChangeEvent event) {
                    maximum_score = event.getValue ();
                    repaintCanvas ();
                  }
                };

              final ScoreChanger score_changer =
                new ScoreChanger ("Score Cutoffs",
                                  minimum_listener, maximum_listener);

              score_changer.setVisible (true);
            }
          });

          canvas.add (popup);

          popup.show (canvas, event.getX (), event.getY ());

        }
      }
    });

    canvas.addMouseMotionListener (new MouseMotionAdapter () {
      public void mouseDragged (final MouseEvent event) {
        updateSelection (event.getPoint ());
        repaintCanvas ();
      }
    });
  }

  /**
   *  Handle a mouse press event on the drawing canvas - select on click,
   *  select and broadcast it on double click.
   **/
  private void handleCanvasMousePress (final MouseEvent event) {
    if (event.getID() != MouseEvent.MOUSE_PRESSED) {
      return;
    }

    if (event.getClickCount () == 2) {
      handleCanvasDoubleClick (event);
    } else {
      handleCanvasSingleClick (event);
    }

    repaintCanvas ();
  }

  /**
   *  Handle a double click on the canvas.
   **/
  private void handleCanvasDoubleClick (final MouseEvent event) {
    if (selected_align_match != null) {
      alignAt (selected_align_match);
    }
  }

  /**
   *  Scroll both the top and bottom
   **/
  private void alignAt (final AlignMatch align_match) {
    owner.unlockDisplays ();

    final int top_sequence_start    = getRealFirstSequenceStart (align_match);
    final int top_sequence_end      = getRealFirstSequenceEnd (align_match);
    final int bottom_sequence_start = getRealSecondSequenceStart (align_match);
    final int bottom_sequence_end   = getRealSecondSequenceEnd (align_match);

    if (top_sequence_start > top_sequence_end &&
        bottom_sequence_end > bottom_sequence_start ||
        top_sequence_start < top_sequence_end &&
        bottom_sequence_end < bottom_sequence_start) {
      try {
        final YesNoDialog yes_no_dialog =
          new YesNoDialog (owner, "reverse and complement bottom sequence?");
        if (yes_no_dialog.getResult ()) {
          owner.getBottomEntryGroup ().reverseComplement ();
        }
      } catch (uk.ac.sanger.pathogens.ReadOnlyException e) {
        // don't bother
      }
    }

    final int new_top_base =
      getRealFirstSequenceStart (align_match) +
      (getRealFirstSequenceEnd (align_match) -
       getRealFirstSequenceStart (align_match)) / 2;
    owner.getTopDisplay ().makeBaseVisible (new_top_base);

    try {
      final Strand top_strand = getTopForwardStrand ();
      final MarkerRange new_top_marker =
        top_strand.makeMarkerRangeFromPositions (top_sequence_start,
                                                 top_sequence_end);
      owner.getTopSelection ().setMarkerRange (new_top_marker);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    final int new_bottom_base =
      getRealSecondSequenceStart (align_match) +
      (getRealSecondSequenceEnd (align_match) -
       getRealSecondSequenceStart (align_match)) / 2;
    owner.getBottomDisplay ().makeBaseVisible (new_bottom_base);

    try {
      final Strand bottom_strand = getBottomForwardStrand ();
      final MarkerRange new_bottom_marker =
        bottom_strand.makeMarkerRangeFromPositions (bottom_sequence_start,
                                                    bottom_sequence_end);
      owner.getBottomSelection ().setMarkerRange (new_bottom_marker);
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    owner.lockDisplays ();
  }

  /**
   *  Handle a single click on the canvas.
   **/
  private void handleCanvasSingleClick (final MouseEvent event) {
    updateSelection (event.getPoint ());
  }

  /**
   *  Set the selection to be the match at the given point.
   **/
  private void updateSelection (final Point point) {
    final AlignMatch clicked_align_match =
      getAlignMatchFromPosition (point);

    selected_align_match = clicked_align_match;
  }

  /**
   *  Return the AlignMatch at the given Point on screen or null if there is
   *  no match at that point.  The alignment_data_array is searched in reverse
   *  order.
   **/
  private AlignMatch getAlignMatchFromPosition (final Point click_point) {
    final int canvas_height = canvas.getSize ().height;
    final int canvas_width = canvas.getSize ().width;

    for (int i = alignment_data_array.length - 1 ; i >= 0  ; --i) {
      final AlignMatch this_match = alignment_data_array [i];

      final int [] match_x_positions =
        getMatchCoords (canvas_width, this_match);

      if (match_x_positions == null) {
        continue;
      }

      if (!checkMatch (this_match)) {
        continue;
      }

      final int top_start_x = match_x_positions[0];
      final int top_end_x = match_x_positions[1];
      final int bottom_start_x = match_x_positions[2];
      final int bottom_end_x = match_x_positions[3];

      // this is the x coordinate of the point where the line y = click_point
      // hits the left edge of the match box
      final double match_left_x =
        top_start_x +
        (1.0 * (bottom_start_x - top_start_x)) *
        (1.0 * click_point.y / canvas_height);

      // this is the x coordinate of the point where the line y = click_point
      // hits the right edge of the match box
      final double match_right_x =
        top_end_x +
        (1.0 * (bottom_end_x - top_end_x)) *
        (1.0 * click_point.y / canvas_height);

//        System.err.println ("match_left_x: " + match_left_x + "  " +
//                            click_point.y);
//        System.err.println ("match_right_x: " + match_right_x);
//        System.err.println ("point: " + click_point);

      if (match_left_x <= click_point.x && match_right_x >= click_point.x ||
          match_left_x >= click_point.x && match_right_x <= click_point.x) {
        return this_match;
      }
    }

    return null;
  }

  /**
   *  Call repaint () on the canvas.
   **/
  private void repaintCanvas () {
    canvas.repaint ();
  }

  /**
   *  The main paint function for the canvas.  An off screen image used for
   *  double buffering when drawing the canvas.
   *  @param g The Graphics object of the canvas.
   **/
  private void paintCanvas (final Graphics g) {
    final int canvas_width = canvas.getSize ().width;
    final int canvas_height = canvas.getSize ().height;

    if (canvas_height <= 0 || canvas_width <= 0) {
      // there is no point drawing into a zero width canvas
      return;
    }

    if (!isVisible ()) {
      return;
    }

    if (font_height == -1) {
      setFontInfo ();
    }

    if (offscreen == null) {
      offscreen = createImage (canvas_width,
                               canvas_height);
    }

    Graphics og = offscreen.getGraphics ();
    og.setClip (0, 0, canvas_width, canvas_height);

    fillBackground (og);

    if (last_top_event != null && last_bottom_event != null) {
      drawAlignments (og);
//      drawLabels (og);
    }

    g.drawImage (offscreen, 0, 0, null);
    og.dispose ();
  }

  /**
   *  Draw the background colour of the frames.
   **/
  private void fillBackground (final Graphics g) {
    g.setColor (Color.white);

    g.fillRect (0, 0, canvas.getSize ().width, canvas.getSize ().height);
  }

  /**
   *  Draw the labels into the given Graphics object.  There is a label at the
   *  top left showing info about the current AlignMatch object, a label at
   *  the bottom left showing whether or not the bottom sequence is reverse
   *  complement and a label beside the Scrollbar showing the current score
   *  cutoff.
   **/
  private void drawLabels (final Graphics g) {
    final FontMetrics fm = g.getFontMetrics ();
    final int canvas_width = canvas.getSize ().width;
    final int canvas_height = canvas.getSize ().height;

    final String cutoff_label =
      Integer.toString (scroll_bar.getValue ());

    final int cutoff_label_width = fm.stringWidth (cutoff_label);

    int cutoff_label_position =
      (int)((scroll_bar.getValue () -
             scroll_bar.getMinimum ()) / (1.0 *
            (scroll_bar.getMaximum () -
             scroll_bar.getMinimum ())) * canvas_height);

    if (cutoff_label_position < font_height) {
      cutoff_label_position = font_height;
    }

    g.drawString (cutoff_label, canvas_width - cutoff_label_width,
                  cutoff_label_position);
  }

  /**
   *  Draw the alignments into the given Graphics object.
   **/
  private void drawAlignments (final Graphics g) {
    final int canvas_height = canvas.getSize ().height;
    final int canvas_width = canvas.getSize ().width;

    for (int i = 0 ; i < alignment_data_array.length ; ++i) {
      final AlignMatch this_match = alignment_data_array [i];

      final int [] match_x_positions =
        getMatchCoords (canvas_width, this_match);

      if (match_x_positions == null) {
        continue;
      }

      final int top_start_x = match_x_positions[0];
      final int top_end_x = match_x_positions[1];
      final int bottom_start_x = match_x_positions[2];
      final int bottom_end_x = match_x_positions[3];

      final int [] x_coords = new int [4];
      final int [] y_coords = new int [4];

      x_coords[0] = top_start_x;
      y_coords[0] = 0;
      x_coords[1] = bottom_start_x;
      y_coords[1] = canvas_height;
      x_coords[2] = bottom_end_x;
      y_coords[2] = canvas_height;
      x_coords[3] = top_end_x;
      y_coords[3] = 0;

      if (!checkMatch (this_match)) {
        continue;
      }

      if (top_end_x - top_start_x < 5 && top_end_x - top_start_x > -5) {
        g.setColor (Color.black);

        g.fillPolygon (x_coords, y_coords, x_coords.length);
      } else {
        if (selected_align_match == this_match) {
          g.setColor (Color.yellow);
        } else {
          final int score = this_match.getScore ();

          if (score == -1) {
            g.setColor (Color.red);
          } else {
            int colour_index = score_colours.length - 1;

            if (maximum_score > minimum_score) {
              colour_index =
                (int)(score_colours.length * 0.999 *
                      (score - minimum_score) /
                      (maximum_score - minimum_score));

            }
            g.setColor (score_colours[colour_index]);
          }
        }

        g.fillPolygon (x_coords, y_coords, x_coords.length);

        g.setColor (Color.black);

        g.drawLine (top_start_x, 0, bottom_start_x, canvas_height);
        g.drawLine (top_end_x, 0, bottom_end_x, canvas_height);
      }
    }
  }

  /**
   *  Return true if and only if the given match is currently visible.
   **/
  private boolean checkMatch (final AlignMatch match) {
    final int score = match.getScore ();

    if (score > -1) {
      if (score < minimum_score || score > maximum_score) {
        return false;
      }
    }

    if (Math.abs (match.getFirstSequenceStart () -
                  match.getFirstSequenceEnd ()) <
        scroll_bar.getValue ()) {
      return false;
    }

    return true;
  }

  /**
   *  Return the screen x positions of the corners of the given match.  The
   *  order is Top Left, Top Right, Bottom Left, Bottom Right, unless the
   *  match is an inversion, in which case it will be TL,TR,BR,BL.  Returns
   *  null if and only if the match is not currently visible.
   **/
  private int [] getMatchCoords (final int canvas_width,
                                 final AlignMatch this_match) {
    final int top_sequence_start    = getRealFirstSequenceStart (this_match);
    final int top_sequence_end      = getRealFirstSequenceEnd (this_match);
    final int bottom_sequence_start = getRealSecondSequenceStart (this_match);
    final int bottom_sequence_end   = getRealSecondSequenceEnd (this_match);

    final int top_start_x =
      getScreenPosition (canvas_width, last_top_event, top_sequence_start);
    final int top_end_x =
      getScreenPosition (canvas_width, last_top_event, top_sequence_end + 1);

    final int bottom_start_x =
      getScreenPosition (canvas_width, last_bottom_event,
                         bottom_sequence_start);
    final int bottom_end_x =
      getScreenPosition (canvas_width, last_bottom_event,
                         bottom_sequence_end + 1);

    boolean top_off_left = false;
    boolean top_off_right = false;
    boolean bottom_off_left = false;
    boolean bottom_off_right = false;

    if (top_start_x < 0 && top_end_x < 0) {
      top_off_left = true;
    }

    if (top_start_x >= canvas_width && top_end_x >= canvas_width) {
      top_off_right = true;
    }

    if (bottom_start_x < 0 && bottom_end_x < 0) {
      bottom_off_left = true;
    }

    if (bottom_start_x >= canvas_width && bottom_end_x >= canvas_width) {
      bottom_off_right = true;
    }

    if ((top_off_left ? 1 : 0) +
        (bottom_off_left ? 1 : 0) +
        (top_off_right ? 1 : 0) +
        (bottom_off_right ? 1 : 0) == 2) {
      return null;
    } else {
      final int [] return_values = new int [4];

      return_values[0] = top_start_x;
      return_values[1] = top_end_x;
      return_values[2] = bottom_start_x;
      return_values[3] = bottom_end_x;

      return return_values;
    }
  }

  /**
   *  Return the current forward Strand of the top EntryGroup.
   **/
  private Strand getTopForwardStrand () {
    return owner.getTopEntryGroup ().getBases ().getForwardStrand ();
  }

  /**
   *  Return the current forward Strand of the bottom EntryGroup.
   **/
  private Strand getBottomForwardStrand () {
    return owner.getBottomEntryGroup ().getBases ().getForwardStrand ();
  }

  /**
   *  Return the start position in the first sequence of the given AlignMatch,
   *  taking into account the current orientation of the sequences.
   **/
  private int getRealFirstSequenceStart (final AlignMatch match) {
    final Strand current_top_forward_strand = getTopForwardStrand ();

    final int top_sequence_start;

    if (top_forward_strand == current_top_forward_strand) {
      return match.getFirstSequenceStart ();
    } else {
      final int tmp_start = match.getFirstSequenceStart ();
      return top_forward_strand.getBases ().getComplementPosition (tmp_start);
    }
  }

  /**
   *  Return the end position in the first sequence of the given AlignMatch,
   *  taking into account the current orientation of the sequences.
   **/
  private int getRealFirstSequenceEnd (final AlignMatch match) {
    final Strand current_top_forward_strand =
      owner.getTopEntryGroup ().getBases ().getForwardStrand ();

    final int top_sequence_end;

    if (top_forward_strand == current_top_forward_strand) {
      return match.getFirstSequenceEnd ();
    } else {
      final int tmp_end = match.getFirstSequenceEnd ();
      return top_forward_strand.getBases ().getComplementPosition (tmp_end);
    }
  }

  /**
   *  Return the start position in the second sequence of the given AlignMatch,
   *  taking into account the current orientation of the sequences.
   **/
  private int getRealSecondSequenceStart (final AlignMatch match) {
    final Strand current_bottom_forward_strand =
      owner.getBottomEntryGroup ().getBases ().getForwardStrand ();

    if (bottom_forward_strand == current_bottom_forward_strand) {
      return match.getSecondSequenceStart ();
    } else {
      final int tmp_start = match.getSecondSequenceStart ();
      return
        bottom_forward_strand.getBases ().getComplementPosition (tmp_start);
    }
  }

  /**
   *  Return the end position in the second sequence of the given AlignMatch,
   *  taking into account the current orientation of the sequences.
   **/
  private int getRealSecondSequenceEnd (final AlignMatch match) {
    final Strand current_bottom_forward_strand =
      owner.getBottomEntryGroup ().getBases ().getForwardStrand ();

    if (bottom_forward_strand == current_bottom_forward_strand) {
      return match.getSecondSequenceEnd ();
    } else {
      final int tmp_end = match.getSecondSequenceEnd ();
      return bottom_forward_strand.getBases ().getComplementPosition (tmp_end);
    }
  }

  /**
   *  Convert a base position into a screen x coordinate.
   **/
  private int getScreenPosition (final int canvas_width,
                                 final DisplayAdjustmentEvent event,
                                 final int base_position) {
    // this is the base that is at the left of the screen
    final int screen_start_base = event.getStart ();

    final double base_pos_float =
      event.getBaseWidth () * (base_position - screen_start_base);

    if (base_pos_float > 30000) {
      return 30000;
    } else {
      if (base_pos_float < -30000) {
        return -30000;
      } else {
        return (int) base_pos_float;
      }
    }
  }

  /**
   *  Set font_width, font_height and font_base_line from this.font
   **/
  private void setFontInfo () {
    FontMetrics fm = getFontMetrics (getFont ());
      //      System.out.println ("fm:" + fm.toString ());

    int [] font_widths = fm.getWidths ();

    font_width = -1;
    // get the maximum character width

    for ( int i = 0 ; i < font_widths.length ; ++i ) {
      if (font_widths[i] > font_width) {
        font_width = font_widths[i];
      }
    }

//      font_width = font_width * 10 / 9;

//      System.out.println ("font width: " + font_width);

    font_height = fm.getHeight ();
    font_base_line = fm.getMaxAscent ();
  }

  /**
   *  Return an array of colours that will be used for colouring the matches
   *  (depending on score).
   **/
  private Color [] makeColours () {
    final Color [] return_colours = {
      new Color (255,255,255),
      new Color (255,239,239),
      new Color (255,223,223),
      new Color (255,207,207),
      new Color (255,191,191),
      new Color (255,175,175),
      new Color (255,159,159),
      new Color (255,143,143),
      new Color (255,127,127),
      new Color (255,111,111),
      new Color (255,95,95),
      new Color (255,79,79),
      new Color (255,63,63),
      new Color (255,47,47),
      new Color (255,31,31),
      new Color (255,15,15),
      new Color (255,0,0),
    };

    return return_colours;
  }

  /**
   *  The reference of the Comparator that owns this component.
   **/
  final Comparator owner;

  /**
   *  The comparison data that will be displayed in this component.
   **/
  final ComparisonData comparison_data;

  /**
   *  The drawing area for this component.
   **/
  private Canvas canvas = null;

  /**
   *  Off screen image used for double buffering when drawing the canvas.
   **/
  private Image offscreen;

  /**
   *  This is the last DisplayAdjustmentEvent reference that was passed to
   *  setTopSeqeuencePosition ().
   **/
  private DisplayAdjustmentEvent last_top_event;

  /**
   *  This is the last DisplayAdjustmentEvent reference that was passed to
   *  setBottomSeqeuencePosition ().
   **/
  private DisplayAdjustmentEvent last_bottom_event;

  /**
   *  Set by the constructor to be the original forward strand for the top
   *  sequence.  This is use to determine whether to top sequence has been
   *  reverse-complemented or not.
   **/
  private Strand top_forward_strand;

  /**
   *  Set by the constructor to be the original forward strand for the bottom
   *  sequence.  This is use to determine whether to bottom sequence has been
   *  reverse-complemented or not.
   **/
  private Strand bottom_forward_strand;

  /**
   *  An array containing all the AlignMatch objects we are displaying.  The
   *  objects are stored in z order.  The last object in the array is drawn
   *  last, so is on top.
   **/
  private AlignMatch [] alignment_data_array;

  /**
   *  This is the last match that the user clicked on.
   **/
  private AlignMatch selected_align_match = null;

  /**
   *  The spectrum of colours use to display the scores of matches.
   **/
  private Color [] score_colours = null;

  /**
   *  The scroll bar used to set the minimum length of the visible matches.
   **/
  private Scrollbar scroll_bar = null;

  /**
   *  Matches with scores below this values will not be shown.
   **/
  private int minimum_score = 0;

  /**
   *  Matches with scores above this values will not be shown.
   **/
  private int maximum_score = 100;

  /**
   *  The height of the font used in this component.
   **/
  private int font_height = -1;

  /**
   *  The (maximum) width of the font used in this component.
   **/
  private int font_width;

  /**
   *  The base line of the font used in this component.
   **/
  private int font_base_line;
}
